# Tai Sakuma <tai.sakuma@gmail.com>
from __future__ import print_function
import time
import logging

import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.concurrently import MultiprocessingDropbox
from alphatwirl.concurrently import TaskPackage

##__________________________________________________________________||
class MockProgressReporter(object):
    pass

@pytest.fixture()
def mock_progressmonitor():
    ret = mock.MagicMock()
    ret.createReporter.return_value = MockProgressReporter()
    return ret

@pytest.fixture()
def obj(mock_progressmonitor):
    ret = MultiprocessingDropbox(progressMonitor=mock_progressmonitor)
    ret.open()
    yield ret
    ret.close()

##__________________________________________________________________||
@pytest.fixture()
def logger():
    ret = logging.getLogger(__name__)
    ret.setLevel(logging.DEBUG)
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    ret.addHandler(handler)
    return ret

def mocktask(self, *args, **kwargs):
        logger = logging.getLogger(__name__)
        logger.info('{}(args={!r}, kwargs={!r})'.format(
            'mocktask', args, kwargs
        ))

@pytest.fixture()
def package1():
    args = (111, 222)
    kwargs = dict(A='abc', B='def')
    return TaskPackage(task=mocktask, args=args, kwargs=kwargs)

##__________________________________________________________________||
def test_put(obj, package1, logger, caplog):
    with caplog.at_level(logging.INFO, logger=__name__):
        obj.put(package1)

    ## caplog doesn't catch probably because of threading
    # assert len(caplog.records) == 1

##__________________________________________________________________||



from .mkdir_p import mkdir_p
from .list_to_aligned_text import list_to_aligned_text
from .quote_string import quote_string
from .profile_func import profile_func, print_profile_func

# to be deleted
from .listToAlignedText import listToAlignedText



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
def merge_in_order(map_, id_, data):

    ## merge to prev
    idx = list(map_.keys()).index(id_)
    idx_prev = idx - 1
    if idx_prev < 0:
        map_[id_] = data
    else:
        data_prev = list(map_.values())[idx_prev]
        id_prev = list(map_.keys())[idx_prev]
        if data_prev is None:
            map_[id_] = data
        else:
            if hasattr(data_prev, 'merge'):
                data_prev.merge(data)
            map_.pop(id_)
            id_ = id_prev
            data = data_prev

    ## merge next
    idx = list(map_.keys()).index(id_)
    idx_next = idx + 1
    if idx_next < len(map_):
        data_next = list(map_.values())[idx_next]
        if data_next is not None:
            if hasattr(data, 'merge'):
                data.merge(data_next)
            map_.pop(list(map_.keys())[idx_next])

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
def linear_search(val, boundaries):
    ret = 0
    for i, low in enumerate(boundaries):
        if low <= val:
            ret = i
        else:
            break
    return ret

##__________________________________________________________________||
def binary_search(val, boundaries):
    min_idx = 0
    max_idx = len(boundaries) - 1
    while min_idx <= max_idx:
        idx = min_idx + (max_idx - min_idx)//2
        low = boundaries[idx]
        if val == low:
            return idx
        up = boundaries[idx + 1]
        if val == up:
            return idx + 1
        if low < val < up:
            return idx
        elif up <= val:
            min_idx = idx
        else: # val < low
            max_idx = idx

##__________________________________________________________________||



import alphatwirl
import unittest

##__________________________________________________________________||
def genFunc():
    yield 101
    yield 102
    yield 103

##__________________________________________________________________||
class IteClass(object):
    def __init__(self):
        self.val = 100
        pass

    def __iter__(self):
        self.val = 101
        yield self
        self.val = 102
        yield self
        self.val = 103
        yield self

##__________________________________________________________________||
class TestExampleGeneratorFunction(unittest.TestCase):

    def test_genFunc_Iteration(self):
        self.assertEqual([101, 102, 103], [e for e in genFunc()])

    def test_genFunc_ManualIteration(self):
        gen = genFunc()
        it = iter(gen)
        self.assertIs(it, gen)
        self.assertEqual(101, next(it))
        self.assertEqual(102, next(it))
        self.assertEqual(103, next(it))
        self.assertRaises(StopIteration, next, it)

##__________________________________________________________________||
class TestExampleIterableObject(unittest.TestCase):

    def test_iteObj_Iteration(self):
        self.assertEqual([101, 102, 103], [i.val for i in IteClass()])

    def test_iteObj_ManualIteration(self):
        itObj = IteClass()
        self.assertEqual(100, itObj.val)
        it = iter(itObj)
        itObj1 = next(it)
        self.assertIs(itObj, itObj1)
        self.assertEqual(101, itObj1.val)
        itObj2 = next(it)
        self.assertIs(itObj, itObj2)
        self.assertEqual(102, itObj2.val)
        itObj3 = next(it)
        self.assertIs(itObj, itObj3)
        self.assertEqual(103, itObj3.val)
        self.assertRaises(StopIteration, next, it)


##__________________________________________________________________||



import unittest

from alphatwirl import quote_string

##__________________________________________________________________||
class TestQuoteString(unittest.TestCase):

    def test_one(self):

        text = 'AAA'
        expected = 'AAA'
        actual = quote_string(text)
        self.assertEqual(expected, actual)

    def test_empty(self):

        text = ''
        expected = '""'
        actual = quote_string(text)
        self.assertEqual(expected, actual)

    def test_space(self):

        text = 'AA A' # space in text
        expected = '"AA A"' # double quoted
        actual = quote_string(text)
        self.assertEqual(expected, actual)

        text = ' AAA' # space at the beginning
        expected = '" AAA"' # double quoted
        actual = quote_string(text)
        self.assertEqual(expected, actual)

        text = 'AAA ' # space at the end
        expected = '"AAA "' # double quoted
        actual = quote_string(text)
        self.assertEqual(expected, actual)

    def test_quoted(self):

        text = '"AAA"' # already double quoted
        expected = r'"\"AAA\""' # escaped and double quoted
        actual = quote_string(text)
        self.assertEqual(expected, actual)

        text = '"AA A"' #
        expected = r'"\"AA A\""' #
        actual = quote_string(text)
        self.assertEqual(expected, actual)

        text = 'AA "ab cde" BB' #
        expected = r'"AA \"ab cde\" BB"' #
        actual = quote_string(text)
        self.assertEqual(expected, actual)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@cern.ch>
from __future__ import print_function
import cProfile, pstats
from io import StringIO, BytesIO

##__________________________________________________________________||
def print_profile_func(func, profile_out_path=None):
    result = profile_func(func)
    if profile_out_path is None:
        print(result)
    else:
        with open(profile_out_path, 'w') as f:
            f.write(result)

##__________________________________________________________________||
def profile_func(func):
    pr = cProfile.Profile()
    pr.enable()
    func()
    pr.disable()
    sortby = 'cumulative'
    try:
        s = StringIO()
        pstats.Stats(pr, stream=s).strip_dirs().sort_stats(sortby).print_stats()
    except TypeError:
        s = BytesIO()
        pstats.Stats(pr, stream=s).strip_dirs().sort_stats(sortby).print_stats()
    return s.getvalue()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import sys
import pytest

from alphatwirl.selection.factories.expand import expand_path_cfg

from alphatwirl.selection.factories.factory import FactoryDispatcher
from alphatwirl.selection.modules.LambdaStr import LambdaStr
from alphatwirl.selection.modules import All, Any, Not

##__________________________________________________________________||
# path_cfg, expanded, obj
params = [
    pytest.param(
        'ev : ev.nJets[0] >= 2',
        dict(
            components=(),
            factory='LambdaStrFactory',
            lambda_str='ev : ev.nJets[0] >= 2',
        ),
        LambdaStr(
            name='ev : ev.nJets[0] >= 2',
            lambda_str='ev : ev.nJets[0] >= 2',
        ),
        id='string:lambda_str'
    ),
    pytest.param(
        'ev : ev.nJets[0] >= {n}',
        dict(
            components=(),
            factory='LambdaStrFactory',
            lambda_str='ev : ev.nJets[0] >= {n}',
        ),
        LambdaStr(
            name='ev : ev.nJets[0] >= 5242',
            lambda_str='ev : ev.nJets[0] >= 5242',
        ),
        id='string:lambda_str-not-formatted'
    ),
    pytest.param(
        dict(All=()),
        {'factory': 'AllFactory', 'components': ()},
        All(name='All', selections=[]),
        id='dict-all-empty'
    ),
    pytest.param(
        dict(Any=()),
        {'factory': 'AnyFactory', 'components': ()},
        Any(name='Any', selections=[]),
        id='dict-any-empty'
    ),
    pytest.param(
        dict(Any=(
            'ev : ev.x[0] == 0',
            dict(All=(
                'ev : ev.x[0] >= 1',
                'ev : ev.y[0] >= 100',
                )),
            dict(Not=dict(
                Any=(
                    'ev : ev.z[0] == 0',
                    'ev : ev.w[0] >= 300',
                ),
            )),
        )),
        dict(
            factory='AnyFactory',
            components=(
                dict(
                    factory='LambdaStrFactory',
                    components=(),
                    lambda_str='ev : ev.x[0] == 0',
                ),
                dict(
                    factory='AllFactory',
                    components=(
                        dict(
                            factory='LambdaStrFactory',
                            components=(),
                            lambda_str='ev : ev.x[0] >= 1',
                        ),
                        dict(
                            factory='LambdaStrFactory',
                            components=(),
                            lambda_str='ev : ev.y[0] >= 100',
                        )
                    )
                ),
                dict(
                    factory='NotFactory',
                    components=(
                        dict(
                            factory='AnyFactory',
                            components=(
                                dict(
                                    factory='LambdaStrFactory',
                                    components=(),
                                    lambda_str='ev : ev.z[0] == 0'
                                ),
                                dict(
                                    factory='LambdaStrFactory',
                                    components=(),
                                    lambda_str='ev : ev.w[0] >= 300',
                                ),
                            ),
                        ),
                    ),
                )
            )
        ),
        Any(
            name='Any',
            selections=[
                LambdaStr(
                    name='ev : ev.x[0] == 0',
                    lambda_str='ev : ev.x[0] == 0'
                ),
                All(
                    name='All',
                    selections=[
                        LambdaStr(
                            name='ev : ev.x[0] >= 1',
                            lambda_str='ev : ev.x[0] >= 1'),
                        LambdaStr(
                            name='ev : ev.y[0] >= 100',
                            lambda_str='ev : ev.y[0] >= 100')
                    ]
                ),
                Not(
                    name='Not',
                    selection=Any(
                        name='Any',
                        selections=[
                            LambdaStr(
                                name='ev : ev.z[0] == 0',
                                lambda_str='ev : ev.z[0] == 0'
                            ),
                            LambdaStr(
                                name='ev : ev.w[0] >= 300',
                                lambda_str='ev : ev.w[0] >= 300'
                            )
                        ]
                    )
                )
            ]
        ),
        id='example',
        ## marks=pytest.mark.skip(reason='not fully expanded')
    ),
]

@pytest.mark.parametrize('path_cfg, expected, _', params)
def test_expand_path_cfg(path_cfg, expected, _):
    actual = expand_path_cfg(path_cfg=path_cfg)
    assert expected == actual

    # give expanded one
    actual = expand_path_cfg(path_cfg=actual)
    assert expected == actual

@pytest.mark.parametrize('path_cfg, _, expected', params)
def test_factory(path_cfg, _, expected):

    kargs = dict(
        AllClass=All, AnyClass=Any, NotClass=Not,
        LambdaStrClass=LambdaStr,
        n=5242,
    )

    obj = FactoryDispatcher(path_cfg=path_cfg, **kargs)
    assert repr(expected) == repr(obj)
    assert str(expected) == str(obj)

##__________________________________________________________________||
@pytest.mark.parametrize('path_cfg, error', [
    pytest.param(
        dict(All=(), Any=()), ValueError, id='multiple-vertices-All-Any'
    ),
    pytest.param(
        dict(All=(), Not=()), ValueError, id='multiple-vertices-All-Not'
    ),
    pytest.param(
        dict(Any=(), Not=()), ValueError, id='multiple-vertices-Any-Not'
    ),
    pytest.param(
        dict(), ValueError, id='empty-dict'
    ),
])
def test_expand_path_cfg_raise(path_cfg, error):
    with pytest.raises(error):
        expand_path_cfg(path_cfg=path_cfg)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import gzip

try:
   import cPickle as pickle
except:
   import pickle

##__________________________________________________________________||
class DatasetLoop(object):

    def __init__(self, datasets, reader):
        self.datasets = datasets
        self.reader = reader

    def __repr__(self):
        name_value_pairs = (
            ('datasets', self.datasets),
            ('reader', self.reader),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def __call__(self):
        self.reader.begin()
        for dataset in self.datasets:
            self.reader.read(dataset)
        return self.reader.end()

##__________________________________________________________________||
class ResumableDatasetLoop(object):

    def __init__(self, datasets, reader, workingarea):
        self.datasets = datasets
        self.reader = reader
        self.workingarea = workingarea

    def __repr__(self):
        name_value_pairs = (
            ('datasets', self.datasets),
            ('reader', self.reader),
            ('workingarea', self.workingarea),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def __call__(self):
        self.reader.begin()
        for dataset in self.datasets:
            self.reader.read(dataset)

        path = os.path.join(self.workingarea.path, 'reader.p.gz')
        with gzip.open(path, 'wb') as f:
            pickle.dump(self.reader, f, protocol=pickle.HIGHEST_PROTOCOL)

        return self.reader.end()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import copy
import logging

##__________________________________________________________________||
N_KEYS = 3
IDX_DEPTH = 0
IDX_PASS = 3
IDX_TOTAL = 4

##__________________________________________________________________||
class Count(object):
    def __init__(self, selections=None):
        if selections is None:
            selections = [ ]

        self._results = [ ]

        for sel in selections:
            self.add(sel)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self._results)

    def copy(self):
        return copy.deepcopy(self)

    def add(self, selection):
        init_result = self._init_result_for_selection(selection)
        self._results.append(init_result)

    def _init_result_for_selection(self, selection):
        #
        class_name = selection.__class__.__name__

        #
        if hasattr(selection, 'name') and selection.name is not None:
            selection_name = selection.name
        else:
            selection_name = ''

        #
        depth = 1
        pass_ = 0
        total = 0

        #
        return [depth, class_name, selection_name, pass_, total]

    def count(self, pass_):
        for r, p in zip(self._results, pass_):
            r[IDX_TOTAL] += 1 # total
            if p: r[IDX_PASS] += 1 # pass

    def increment_depth(self, by = 1):
        for r in self._results:
            r[IDX_DEPTH] += by

    def insert(self, i, other):
        self._results[(i + 1):(i + 1)] = other._results

    def results(self):
        return self._results

    def __add__(self, other):
        ret = self.copy()

        if other == 0: # other is 0 when e.g. sum([obj1, obj2])
            return ret

        self._add_results_inplace(ret._results, other._results)
        return ret

    def __iadd__(self, other):
        self._add_results_inplace(self._results, other._results)
        return self

    def __radd__(self, other):
        return self.__add__(other)

    def _add_results_inplace(self, res1, res2):
        if not len(res1) == len(res2):
            logger = logging.getLogger(__name__)
            logger.warning('cannot add because res1 and res2 don\'t have the same length: res1 = {}, res2 = {}'.format(res1, res2))
            return

        if not all([(r1[:N_KEYS] == r2[:N_KEYS]) for r1, r2 in zip(res1, res2)]):
            logger = logging.getLogger(__name__)
            logger.warning('cannot add because res1 and res2 don\'t have the same key columns: res1 = {}, res2 = {}'.format(res1, res2))
            return

        for r1, r2 in zip(res1, res2):
            r1[IDX_PASS] += r2[IDX_PASS]
            r1[IDX_TOTAL] += r2[IDX_TOTAL]

    def to_tuple_list(self):
        return [tuple(e) for e in self._results]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import time
import collections


# This file is stored here rather than in tests/ so that it can be
# found by pickle.load()

##__________________________________________________________________||
MockResult = collections.namedtuple('MockResult', 'name args kwargs')

class MockTask(object):
    def __init__(self, name, time):
        self.name = name
        self.time = time

    def __call__(self, *args, **kwargs):
        time.sleep(self.time)
        return MockResult(name=self.name, args=args, kwargs=kwargs)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
def create_files_start_length_list(
        files, func_get_nevents_in_file=None,
        max_events=-1, max_events_per_run=-1,
        max_files=-1, max_files_per_run=1):

    files = _apply_max_files(files, max_files)

    if max_events == 0 or max_events_per_run == 0:
        return [ ]

    if max_events < 0 and max_events_per_run < 0:
        return _fast_path(files, max_files_per_run)

    return _full_path(files, func_get_nevents_in_file, max_events,
                           max_events_per_run, max_files_per_run)

##__________________________________________________________________||
def _apply_max_files(files, max_files):
    if max_files < 0:
        return files
    return files[:min(max_files, len(files))]

def _fast_path(files, max_files_per_run):
    if not files:
        return [ ]
    if max_files_per_run < 0:
        return [(files, 0, -1)]
    if max_files_per_run == 0:
        return [ ]
    return [(files[i:(i + max_files_per_run)], 0, -1) for i in range(0, len(files), max_files_per_run)]

def _full_path(files, func_get_nevents_in_file, max_events, max_events_per_run, max_files_per_run):

    if max_events == 0 or max_events_per_run == 0 or max_files_per_run == 0:
        return [ ]

    # this can be slow
    file_nevents_list = _file_nevents_list(
        files,
        func_get_nevents_in_file=func_get_nevents_in_file,
        max_events=max_events
    )

    file_nevents_list = _apply_max_events_total(
        file_nevents_list, max_events
    )

    files_start_length_list = _files_start_length_list(
        file_nevents_list, max_events_per_run, max_files_per_run
    )

    return files_start_length_list

def _file_nevents_list(files, func_get_nevents_in_file, max_events):
    total_events = 0
    ret = [ ]
    for f in files:
        if 0 <= max_events <= total_events:
            break

        # this can be slow
        n = func_get_nevents_in_file(f)

        if n is None:
            continue

        if n == 0:
            continue

        ret.append((f, n))
        total_events += n
    return ret

def _apply_max_events_total(file_nevents_list, max_events_total):

    if max_events_total < 0:
        return file_nevents_list

    ret = [ ]
    for file, nevents in file_nevents_list:
        if max_events_total == 0:
            break
        nevents = min(max_events_total, nevents)
        ret.append((file, nevents))
        max_events_total -= nevents
    return ret

##__________________________________________________________________||
def _files_start_length_list(file_nevents_list, max_events_per_run, max_files_per_run):

    if not file_nevents_list:
        return [ ]

    if max_events_per_run == 0 or max_files_per_run == 0:
        return [ ]

    total_nevents = sum([n for f, n, in file_nevents_list])
    if total_nevents == 0:
        return [ ]

    if max_events_per_run < 0:
        max_events_per_run = total_nevents

    total_nfiles = len(set([f for f, n, in file_nevents_list]))
    if max_files_per_run < 0:
        max_files_per_run = total_nfiles

    ##
    files = [ ]
    nevents = [ ]
    start = [ ]
    length = [ ]
    i = 0
    for file_, nev in file_nevents_list:

        if nev == 0:
            continue

        if i == len(files):
            # create a new run
            files.append([ ])
            nevents.append(0)
            start.append(0)
            length.append(0)

        files[i].append(file_)
        nevents[i] += nev

        if max_events_per_run >= nevents[i]:
            length[i] = nevents[i]

        else:
            dlength = max_events_per_run - length[i]
            length[i] = max_events_per_run

            i += 1
            files.append([file_])
            nevents.append(nevents[i-1] - length[i-1])
            start.append(dlength)

            while max_events_per_run < nevents[i]:
                length.append(max_events_per_run)

                i += 1
                files.append([file_])
                nevents.append(nevents[i-1] - length[i-1])
                start.append(start[i-1] + length[i-1])

            length.append(nevents[i])

        if max_events_per_run == nevents[i]:
            i += 1 # to next run
            continue

        if max_files_per_run == len(files[i]):
            i += 1 # to next run

    return list(zip(files, start, length))

##__________________________________________________________________||



import unittest

##__________________________________________________________________||
hasROOT = False
try:
    import ROOT
    hasROOT = True
except ImportError:
    pass

if hasROOT:
    from alphatwirl.roottree import BranchAddressManager

##__________________________________________________________________||
class MockLeaf(object):
    def __init__(self, name, typename, leafcount = None, maximum = None):
        self.name = name
        self.typename = typename
        self.leafcount = leafcount
        self.maximum = maximum
    def GetName(self): return self.name
    def GetTypeName(self): return self.typename
    def GetLeafCount(self): return self.leafcount
    def GetMaximum(self): return self.maximum

##__________________________________________________________________||
class MockNullLeaf(object):
    def GetName(self): raise ReferenceError('null object')

##__________________________________________________________________||
class MockTree(object):

    def __init__(self, leaves):
        self.leaves = leaves
        self.leafDict = dict([(l.GetName(), l) for l in leaves])
        self.branchstatus = [ ]
        self.branchaddress = [ ]

    def GetListOfLeaves(self):
        return self.leaves

    def GetLeaf(self, name):
        return self.leafDict[name]

    def SetBranchStatus(self, name, status):
        self.branchstatus.append((name, status))

    def SetBranchAddress(self, name, address):
        self.branchaddress.append((name, address))

##__________________________________________________________________||
@unittest.skipUnless(hasROOT, "has no ROOT")
class TestBranchAddressManager(unittest.TestCase):


    def setUp(self):

        run = MockLeaf('run', 'UInt_t', MockNullLeaf())
        lumi = MockLeaf('lumi', 'UInt_t', MockNullLeaf())
        evt  = MockLeaf('evt', 'Int_t', MockNullLeaf())
        met_pt = MockLeaf('met_pt', 'Double_t', MockNullLeaf())
        njet = MockLeaf('njet', 'Int_t', MockNullLeaf(), 13)
        jet_pt = MockLeaf('jet_pt', 'Double_t', njet)
        nele = MockLeaf('nele', 'Int_t', MockNullLeaf(), 0)
        ele_pt = MockLeaf('ele_pt', 'Double_t', nele)
        self.tree1 = MockTree(leaves = [run, lumi, evt, met_pt, njet, jet_pt, nele, ele_pt])

    def tearDown(self):

        BranchAddressManager.arrayDict.clear()
        BranchAddressManager.counterArrayDict.clear()

    def test_getArrays(self):

        manager = BranchAddressManager()

        array_jet_pt, array_jet_pt_count = manager.getArrays(self.tree1, 'jet_pt')
        self.assertEqual('d', array_jet_pt.typecode)
        self.assertEqual(13, array_jet_pt.buffer_info()[1])
        self.assertEqual('i', array_jet_pt_count.typecode)
        self.assertEqual(1, array_jet_pt_count.buffer_info()[1])

    def test_getArrays_same_objects(self):

        manager = BranchAddressManager()

        array_jet_pt_1, array_jet_pt_count_1 = manager.getArrays(self.tree1, 'jet_pt')
        array_jet_pt_2, array_jet_pt_count_2 = manager.getArrays(self.tree1, 'jet_pt')
        self.assertIs(array_jet_pt_1, array_jet_pt_2)
        self.assertIs(array_jet_pt_count_1, array_jet_pt_count_2)

    def test_getArrays_same_objects_different_managers(self):

        manager1 = BranchAddressManager()
        manager2 = BranchAddressManager()

        array_jet_pt_1, array_jet_pt_count_1 = manager1.getArrays(self.tree1, 'jet_pt')
        array_jet_pt_2, array_jet_pt_count_2 = manager2.getArrays(self.tree1, 'jet_pt')
        self.assertIs(array_jet_pt_1, array_jet_pt_2)
        self.assertIs(array_jet_pt_count_1, array_jet_pt_count_2)

    def test_getArrays_no_count(self):

        manager = BranchAddressManager()

        array_met_pt, array_met_pt_count = manager.getArrays(self.tree1, 'met_pt')
        self.assertEqual('d', array_met_pt.typecode)
        self.assertEqual(1, array_met_pt.buffer_info()[1])
        self.assertIsNone(array_met_pt_count)

    def test_getArrays_no_branch(self):

        manager = BranchAddressManager()

        array_zet_pt, array_zet_pt_count = manager.getArrays(self.tree1, 'zet_pt')
        self.assertIsNone(array_zet_pt)
        self.assertIsNone(array_zet_pt_count)

    def test_getArrays_zero_max(self):

        manager = BranchAddressManager()

        array_ele_pt, array_ele_pt_count = manager.getArrays(self.tree1, 'ele_pt')
        self.assertEqual('d', array_ele_pt.typecode)
        self.assertEqual(1, array_ele_pt.buffer_info()[1])
        self.assertEqual('i', array_ele_pt_count.typecode)
        self.assertEqual(1, array_ele_pt_count.buffer_info()[1])

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

from .quote_string import quote_string

##__________________________________________________________________||
def list_to_aligned_text(src, format_dict = None, left_align_last_column = False):

    # e.g.,
    # src = [
    #     ('component', 'v1', 'nvar', 'n'),
    #     ('data1',  100, 6.0,   40),
    #     ('data1',    2, 9.0, 3.3),
    #     ('data1', 3124, 3.0, 0.0000001),
    #     ('data2',  333, 6.0, 300909234),
    #     ('data2',   11, 2.0, 323432.2234),
    # ]

    if not src: return '' # src = [ ]
    if not src[0]: return '' # e.g., src = [(), (), ()]

    transposed = [[r[i] for r in src] for i in range(len(src[0]))]
    # e.g.,
    # transposed = [
    #     ['component', 'data1', 'data1', 'data1', 'data2', 'data2'],
    #     ['v1', 100, 2, 3124, 333, 11],
    #     ['nvar', 6.0, 9.0, 3.0, 6.0, 2.0],
    #     ['n', 40, 3.3, 1e-07, 300909234, 323432.2234],
    #     ]

    if format_dict is not None:
    # format_dict = dict(n = '{:10.2f}')

        for column in transposed:

            colum_name = column[0]
            if colum_name not in format_dict:
                continue

            colum_format = format_dict[colum_name]
            column[:] = [column[0]] + [colum_format.format(c) for c in column[1:]]
            # e.g., .['n', '40.00', '3.30', '0.00', '300909234.00', '323432.22']

    # https://stackoverflow.com/questions/25898733/why-does-strfloat-return-more-digits-in-python-3-than-python-2
    # str(0.41999999999999993) is 0.42 in python2 but is 0.41999999999999993 in python3
    # use this format to make the output in python 3 as in python 2
    default_float_format = '{:.12g}'

    transposed = [[str(int(e)) if isinstance(e, float) and e.is_integer() else e for e in r] for r in transposed]
    transposed = [[default_float_format.format(e) if isinstance(e, float) else e for e in r] for r in transposed]
    transposed = [[str(e) for e in r] for r in transposed]

    transposed = [[quote_string(e) for e in r] for r in transposed]

    columnWidths = [max([len(e) for e in r]) for r in transposed]
    # e.g., columnWidths = [9, 2, 4, 1]

    formatList = ['{:>' + str(e) + 's}' for e in columnWidths]
    # e.g., formatList = ['{:>9s}', '{:>4s}', '{:>4s}', '{:>11s}']

    if left_align_last_column:
        formatList[-1] = '{}'

    format = " " + " ".join(formatList)
    # e.g., format = "{:>9s} {:>4s} {:>4s} {:>11s}"

    ret = "\n".join([format.format(*row) for row in zip(*transposed)]) + "\n"
    # example ret
    # component   v1 nvar           n
    #     data1  100    6          40
    #     data1    2    9         3.3
    #     data1 3124    3       1e-07
    #     data2  333    6   300909234
    #     data2   11    2 323432.2234

    return ret

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from __future__ import print_function
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.selection.modules import All, Any, Not
from alphatwirl.selection.modules import AllwCount, AnywCount, NotwCount

##__________________________________________________________________||
all_classes = [All, AllwCount]
all_classe_ids = [c.__name__ for c in all_classes]

any_classes = [Any, AnywCount]
any_classe_ids = [c.__name__ for c in any_classes]

not_classes = [Not, NotwCount]
not_classe_ids = [c.__name__ for c in not_classes]

##__________________________________________________________________||
allany_classes = all_classes + any_classes
allany_classe_ids = all_classe_ids + any_classe_ids

##__________________________________________________________________||
@pytest.fixture()
def mocksel():
    ret = mock.MagicMock()
    ret.__str__.return_value = '\n'.join(['Mocksel:', '    line1', '    line2'])
    return ret

##__________________________________________________________________||
@pytest.mark.parametrize('Class', allany_classes, ids=allany_classe_ids)
def test_allany_str(Class, mocksel):
    obj1 = Class(selections=[mocksel, mocksel])
    obj0 = Class(selections=[obj1])
    str(obj1)
    str(obj0)

##__________________________________________________________________||



from .Events import Events
from .Branch import Branch
from .EventBuilderConfig import EventBuilderConfig

hasROOT = False
try:
    import ROOT
    hasROOT = True
except ImportError:
    pass

if hasROOT:
    from .build import BuildEvents
    from .BEvents import BEvents
    from .BEventBuilder import BEventBuilder
    from .BranchBuilder import BranchBuilder
    from .EventBuilder import EventBuilder
    from .BranchAddressManager import BranchAddressManager
    from .BranchAddressManagerForVector import BranchAddressManagerForVector
    from .inspect import inspect_tree



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import sys
import pytest

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True

if not has_no_ROOT:
    from alphatwirl.roottree import BEvents as Events
    from alphatwirl.roottree import Branch

##__________________________________________________________________||
pytestmark = pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")

##__________________________________________________________________||
@pytest.fixture()
def tree():
    input_file_name = 'sample_02.root'
    tree_name = 'tree'
    input_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', input_file_name)

    input_file = ROOT.TFile.Open(input_path)
    yield input_file.Get(tree_name)

@pytest.fixture()
def events(tree):
    yield Events(tree)

@pytest.mark.skipif(sys.version_info[0]!=2, reason="skip for Python 3")
def test_vector(tree, events):
    trigger_path = events.trigger_path
    trigger_decision = events.trigger_decision

    assert tree.GetEntry(0) > 0
    assert 449 == len(trigger_path)
    assert 'AlCa_EcalEtaEBonly' == trigger_path[0]
    assert 'DST_Physics' == trigger_path[12]
    assert 'HLT_SingleForJet25' == trigger_path[438]
    assert 449 == len(trigger_decision)
    assert 0 == trigger_decision[0]
    assert 0 == trigger_decision[13]
    assert 0 == trigger_decision[438]

    assert tree.GetEntry(1) > 0
    assert 438 == len(trigger_path)
    assert 'AlCa_EcalEtaEBonly' == trigger_path[0]
    assert 'DST_Ele8_CaloIdL_CaloIsoVL_TrkIdVL_TrkIsoVL_HT250' == trigger_path[12]
    with pytest.raises(IndexError):
        trigger_path[438]
    assert 438 == len(trigger_decision)
    assert 0 == trigger_decision[0]
    assert 1 == trigger_decision[13]
    with pytest.raises(IndexError):
        trigger_decision[438]

    # This sample file has only two entries. When the 3rd entry is
    # tried to be accessed, GetEntry(2) returns 0, but the vectors
    # won't be cleared. These have the previous contents.
    assert tree.GetEntry(2) == 0
    assert 438 == len(trigger_path)
    assert 438 == len(trigger_decision)

##__________________________________________________________________||



import unittest

from .mock import MockReader, MockSummarizer

##__________________________________________________________________||
class TestMockReader(unittest.TestCase):

    def setUp(self):
        self.summarizer = MockSummarizer([])
        self.obj = MockReader(self.summarizer)

    def tearDown(self):
        pass

    def test_repr(self):
        repr(self.obj)

    def test_results(self):
        self.assertIs(self.summarizer, self.obj.results())

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging

import ROOT

from .inspect import is_ROOT_null_pointer

##__________________________________________________________________||
class BuildEvents(object):
    def __init__(self, config):
        self.config = config

    def __repr__(self):
        name_value_pairs = (
            ('config', self.config),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def __call__(self):
        paths = self.config['file_paths']
        if self.config['check_files']:
            paths = self._verify_files(paths, self.config['skip_error_files'])
        chain = ROOT.TChain(self.config['tree_name'])
        for path in paths:
            chain.Add(path)
        events = self.config['events_class'](
            chain, self.config['max_events'], self.config['start'])
        events.config = self.config
        return events

    def _verify_files(self, paths, skip_error_files):
        ret = [ ]
        for path in paths:
            file_ = ROOT.TFile.Open(path)
            if is_ROOT_null_pointer(file_) or file_.IsZombie():
                logger = logging.getLogger(__name__)
                if skip_error_files:
                    logger.warning('cannot open {}'.format(path))
                    continue
                logger.error('cannot open {}'.format(path))
                raise OSError('cannot open {}'.format(path))
            ret.append(path)
        return ret

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class Combine(object):
    """A combine two binnings.

    This class uses *low* for the values below *at* and otherwise uses
    *high*.

    This class requires bin labels of both binnings to be values in
    the bins.

    """
    def __init__(self, low, high, at):
        self._low = low
        self._high = high
        self._at = at

    def __repr__(self):
        return '{}(low={!r}, high={!r}), at={!r})'.format(
            self.__class__.__name__,
            self._low, self._high, self._at
        )

    def __call__(self, val):
        if val < self._at:
            return self._low(val)
        else:
            return self._high(val)

    def next(self, bin):
        if bin < self._at:
            bin = self._low.next(bin)
            if bin < self._at:
                return bin
            else:
                return self._high(bin)
        return self._high.next(bin)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import gzip
import pytest

try:
    import cPickle as pickle
except:
    import pickle

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.datasetloop import ResumableDatasetLoop

##__________________________________________________________________||
class MockReader(object):
    def begin(self):
        pass

    def read(self, dataset):
        pass

    def end(self):
        pass

##__________________________________________________________________||

@pytest.fixture()
def reader():
    ret = MockReader()
    ret.original_id = id(ret)
    return ret

@pytest.fixture()
def datasets():
    dataset1 = mock.Mock()
    dataset2 = mock.Mock()
    return [dataset1, dataset2]

@pytest.fixture()
def workingarea(tmpdir_factory):
    ret = mock.Mock()
    ret.path = str(tmpdir_factory.mktemp(''))
    return ret

@pytest.fixture()
def obj(datasets, reader, workingarea):
    return ResumableDatasetLoop(datasets, reader, workingarea)

def test_repr(obj):
    repr(obj)

def test_call(obj, reader, workingarea):
    result = obj()
    path = os.path.join(workingarea.path, 'reader.p.gz')
    with gzip.open(path, 'rb') as f:
        reader_unpickled = pickle.load(f)
    assert reader.original_id == reader_unpickled.original_id

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import numpy as np
import functools
import pytest

from alphatwirl.binning import RoundLog

##__________________________________________________________________||
def test_repr():
    obj = RoundLog()
    repr(obj)

def test_default():
    obj = RoundLog()
    assert 1.9952623149688 == pytest.approx(obj(   2))
    assert 19.952623149688 == pytest.approx(obj(  20))
    assert 199.52623149688 == pytest.approx(obj( 200))

def test_next():
    obj = RoundLog()
    assert 2.51188643150958 == pytest.approx(obj.next(2.23872113856834))
    assert 25.11886431509581 == pytest.approx(obj.next(22.3872113856834))
    assert 251.18864315095848 == pytest.approx(obj.next(223.872113856834))

@pytest.mark.parametrize('min_', [None, 10])
@pytest.mark.parametrize('underflow_bin', [None, -1, 0, 0.001])
def test_zero(min_, underflow_bin):
    obj = RoundLog(0.1, 100, min_, underflow_bin)
    if min_ is None:
        assert 0 == obj(0)
        assert 0 == obj(-0)
        assert 0 == obj(0.0)
        assert 0 == obj(-0.0)
    else:
        assert obj(0) is underflow_bin
        assert obj(-0) is underflow_bin
        assert obj(0.0) is underflow_bin
        assert obj(-0.0) is underflow_bin

    if min_ is None:
        assert 0 == obj.next(0) # next to 0 is 0 unless 0 is the
                                # underflow bin
    else:
        if underflow_bin is None:
            assert obj.next(underflow_bin) is None
        else:
            assert obj(min_) == obj.next(underflow_bin)
            # the next to the underflow bin is the bin for the min

@pytest.mark.parametrize('min_', [None, 10])
@pytest.mark.parametrize('underflow_bin', [None, -1, 0, 0.001])
def test_negative(min_, underflow_bin):
    obj = RoundLog(0.1, 100, min_, underflow_bin)
    if min_ is None:
        assert obj(-2.1) is None
        assert obj(float('-inf')) is None
    else:
        assert obj(-2.1) is underflow_bin
        assert obj(float('-inf')) is underflow_bin

def test_valid():
    obj = RoundLog(valid=lambda x: x >= 10)
    assert 12.589254117941675 == pytest.approx(obj(13))
    assert 10.0 == pytest.approx(obj(10))
    assert obj(7) is None

def test_on_boundary():
    obj = RoundLog(0.1, 100)
    assert 100 == obj(100)

@pytest.mark.parametrize('width', [0.1, 0.2])
@pytest.mark.parametrize('underflow_bin', [None, -1, 0, 0.001])
def test_min_on_a_boundary(width, underflow_bin):
    # this test is related to
    # the issue 43
    # https://github.com/alphatwirl/alphatwirl/issues/43

    min_ = 10 # on a boundary
    obj = RoundLog(width, 100, min=min_, underflow_bin=underflow_bin)

    # boundaries (on a computer)
    # when width = 0.1:
    #   [10.00, 12.59, 15.85, 19.95, 25.12, 31.62,
    #    39.81, 50.12, 63.10, 79.43, 100.00]
    #   10.00 is actually 9.999999999999982, so is the first bin
    #
    # when width = 0.2:
    #   [6.31, 10.00, 15.85, 25.12, 39.81, 63.10, 100.00']
    #   10.00 here actually is 10.000000000000005, so is the 2nd bin

    assert 100 == obj(100)
    # 100 is exact because it is the given boundary

    # min_=10 is a boundary, but not necessarily exact.
    if min_ == pytest.approx(obj(min_)):
        # when width = 0.1 (in the above example)
        assert obj(11) == obj(min_)
        assert obj(9) is underflow_bin
    else:
        # when width = 0.2 (in the above example)
        assert obj(9) is not underflow_bin
        assert obj(9) == obj(min_)

    assert obj(6.3) is underflow_bin

    if underflow_bin is None:
        assert obj.next(underflow_bin) is None
    else:
        assert obj(min_) == obj.next(underflow_bin)
        # the next to the underflow bin is the bin for the min

def test_max():
    obj = RoundLog(0.1, 100, max=900)
    # max=900 is not a boundary

    assert 100 == obj(100)

    assert obj(900) is not None
    assert obj(1100) is None

@pytest.mark.parametrize('width', [0.1, 0.2])
@pytest.mark.parametrize('log10_max', [1.6, 2.4])
@pytest.mark.parametrize('overflow_bin', [None, True, 1000])
def test_max_float_on_a_boundary(width, log10_max, overflow_bin):

    max_ = 10**log10_max # max_ on a boundary (within rounding of float)
    obj = RoundLog(width, 10.0, max=max_, overflow_bin=overflow_bin)

    log10_boundaries = obj._round.boundaries

    if overflow_bin is True:
        overflow_bin = 10**log10_boundaries[-1]

    assert 10.0 == obj(10.0)
    # this is always true because it is the given boundary.

    # If the max is exactly a boundary, the max is the upper edge of
    # the last bin and bin(max) is overflow because the upper edge is
    # open. Otherwise, bin(max) is in the last bin.
    if log10_max == log10_boundaries[-1]:
        assert obj(max_) == overflow_bin
    else:
        assert 10**log10_boundaries[-2] == obj(max_)

    # The max is either the upper or lower edge of the last bin, but
    # not necessarily exact.
    if not log10_max == pytest.approx(log10_boundaries[-1]):
        assert log10_max == pytest.approx(log10_boundaries[-2])

    # the next to the last bin is the overflow bin
    assert obj.next(10**log10_boundaries[-2]) == overflow_bin

    # the next to the overflow bin is the overflow bin
    assert obj.next(overflow_bin) == overflow_bin

@pytest.mark.parametrize('max_', [None, 1000])
@pytest.mark.parametrize('overflow_bin', [None, True, 1000])
def test_inf(max_, overflow_bin):
    obj = RoundLog(0.1, 10.0, max=max_, overflow_bin=overflow_bin)

    log10_boundaries = obj._round.boundaries
    if overflow_bin is True:
        overflow_bin = 10**log10_boundaries[-1]

    if max_ is None:
        assert obj(float('inf')) is None
        assert obj(float('-inf')) is None
        assert obj.next(float('inf')) is None
        assert obj.next(float('-inf')) is None
    else:
        assert obj(float('inf')) == overflow_bin
        assert obj(float('-inf')) is None

##__________________________________________________________________||
def to_be_benchmarked(obj):
    val = np.random.exponential(scale=100)
    obj(val)

@pytest.mark.skip(reason='for optimizing for speed')
def test_benchmark(benchmark):
    np.random.seed(0)
    obj = RoundLog(0.1, 100)
    benchmark(to_be_benchmarked, obj)

##__________________________________________________________________||
def to_be_profiled(obj, vals):
    for val in vals:
        obj(val)

@pytest.mark.skip(reason='for optimizing for speed')
def test_profile():
    np.random.seed(0)
    from alphatwirl.misc import profile_func
    obj = RoundLog(0.05, 100)
    vals = np.random.exponential(scale=100, size=500000)
    print(profile_func(functools.partial(to_be_profiled, obj, vals)))

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys

import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.progressbar import Queue, ProgressMonitor
from alphatwirl import progressbar

##__________________________________________________________________||
@pytest.fixture()
def presentation():
    return mock.MagicMock()

@pytest.fixture()
def queue(presentation):
    return Queue(presentation)

@pytest.fixture()
def report():
    return mock.MagicMock()

##__________________________________________________________________||
def test_queue_put(queue, presentation, report):
    queue.put(report)
    presentation.present.assert_called_once_with(report)

##__________________________________________________________________||
@pytest.fixture()
def mock_queue():
    return mock.MagicMock()

@pytest.fixture()
def MockQueue(monkeypatch, mock_queue):
    module = sys.modules['alphatwirl.progressbar.ProgressMonitor']
    ret = mock.MagicMock()
    monkeypatch.setattr(module, 'Queue', ret)
    ret.return_value = mock_queue
    return ret

@pytest.fixture()
def mock_reporter():
    return mock.MagicMock()

@pytest.fixture()
def MockReporter(monkeypatch, mock_reporter):
    module = sys.modules['alphatwirl.progressbar.ProgressMonitor']
    ret = mock.MagicMock()
    monkeypatch.setattr(module, 'ProgressReporter', ret)
    ret.return_value = mock_reporter
    return ret

@pytest.fixture()
def monitor(presentation, MockReporter, MockQueue):
    return ProgressMonitor(presentation)

##__________________________________________________________________||
def test_init(monitor, mock_queue, MockQueue, presentation):
    assert [mock.call(presentation = presentation)] == MockQueue.call_args_list
    assert monitor.queue is mock_queue

def test_begin_end(monitor, MockReporter):
    monitor.begin()
    assert progressbar._progress_reporter is MockReporter()
    monitor.end()
    assert progressbar._progress_reporter is None

def test_createReporter(monitor, mock_reporter, MockReporter, presentation):
    reporter = monitor.createReporter()
    assert [mock.call(queue = monitor.queue)] == MockReporter.call_args_list
    assert reporter is mock_reporter

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import logging
import collections
import gzip

import pytest

try:
    import cPickle as pickle
except:
    import pickle

from alphatwirl.concurrently import WorkingArea
from alphatwirl import mkdir_p

##__________________________________________________________________||
## cannot be replaced with mock.Mock because mock.Mock is not picklable
MockPackage = collections.namedtuple('MockPackage', 'name')
MockResult = collections.namedtuple('MockResult', 'name')

##__________________________________________________________________||
@pytest.fixture()
def topdir(tmpdir_factory):
    ret = str(tmpdir_factory.mktemp(''))
    ret = os.path.join(ret, '_ccsp_temp')
    return ret

##__________________________________________________________________||
def test_renamed(topdir, caplog):
    with caplog.at_level(logging.WARNING):
        obj = WorkingArea(dir=topdir, python_modules=('alphatwirl', ))

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'WorkingArea' in caplog.records[0].name
    assert 'renamed' in caplog.records[0].msg

##__________________________________________________________________||
@pytest.fixture()
def obj(topdir):
    return WorkingArea(topdir=topdir, python_modules=('alphatwirl', ))

##__________________________________________________________________||
def test_repr(obj):
    repr(obj)

def test_open(obj):
    assert obj.path is None

    obj.open()
    assert obj.path is not None
    assert -1 == obj.last_package_index
    assert os.path.isdir(obj.path)
    assert os.path.isfile(os.path.join(obj.path, 'run.py'))
    assert os.path.isfile(os.path.join(obj.path, 'logging_levels.json.gz'))
    assert os.path.isfile(os.path.join(obj.path, 'python_modules.tar.gz'))

def test_query(obj):
    obj.open()
    assert 'run.py' == obj.executable
    assert set([
        'python_modules.tar.gz',
        'logging_levels.json.gz']) == obj.extra_input_files

def test_query_no_python_module(topdir):
    obj = WorkingArea(topdir=topdir)
    obj.open()
    assert 'run.py' == obj.executable
    assert set(['logging_levels.json.gz']) == obj.extra_input_files

def test_put_package(obj):

    obj.open()

    package1 = MockPackage(name='package1')
    package_index = obj.put_package(package1)
    package_path = obj.package_path(package_index)
    package_fullpath = os.path.join(obj.path, package_path)
    assert os.path.isfile(package_fullpath)
    with gzip.open(package_fullpath, 'rb') as f:
       assert package1 == pickle.load(f)

def test_collect_result(obj):

    obj.open()

    result = MockResult(name='result1')

    package_index = 9
    dirname = 'task_{:05d}'.format(package_index)
    result_dir = os.path.join(obj.path, 'results', dirname)
    mkdir_p(result_dir)
    result_path = os.path.join(result_dir, 'result.p.gz')
    with gzip.open(result_path, 'wb') as f:
       pickle.dump(result, f)
       f.close()

    assert result == obj.collect_result(package_index=package_index)

def test_collect_result_ioerror(obj, caplog):
   # the file 'result.p.gz' doesn't exist
   # gzip.open() raises IOFError

    obj.open()

    package_index = 9
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        ret = obj.collect_result(package_index=package_index)

    assert ret is None

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'WorkingArea' in caplog.records[0].name
    # assert 'No such file or directory' in caplog.records[0].msg

def test_collect_result_eoferror(obj):
   # the file 'result.p.gz' is empty.
   # pickle.load() raises EOFError

    obj.open()

    package_index = 9
    dirname = 'task_{:05d}'.format(package_index)
    result_dir = os.path.join(obj.path, 'results', dirname)
    mkdir_p(result_dir)

    result_path = os.path.join(result_dir, 'result.p.gz')
    with open(result_path, 'wb') as f:
       f.close()

    assert obj.collect_result(package_index=package_index) is None

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import itertools

##__________________________________________________________________||
class Base(object):
    """The base class of the classes All and Any

    """

    def __init__(self, name, selections):
        self.name = name
        self.selections = list(selections)

        self._repr_pairs = [
            ('name', self.name),
            ('selections', self.selections),
        ]

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in self._repr_pairs]),
        )

    def __str__(self):
        nwidth = max(len(n) for n, _ in self._repr_pairs)
        nwidth += 4
        nwidth_nested = 12
        lines = [
            '{}:'.format(self.__class__.__name__),
            '{:>{}}: {!r}'.format('name', nwidth, self.name),
            '{:>{}}:'.format('selections', nwidth),
        ]
        nested_lines = list(itertools.chain(*[str(s).split('\n') for s in self.selections]))
        lines.extend(
            ['{}{}'.format(' '*nwidth_nested, str(l)) for l in nested_lines]
        )
        return '\n'.join(lines)

    def add(self, selection):
        self.selections.append(selection)

    def begin(self, event):
        for s in self.selections:
            if hasattr(s, 'begin'): s.begin(event)

    def event(self, event):
        return self(event)

    def end(self):
        for s in self.selections:
            if hasattr(s, 'end'): s.end()

##__________________________________________________________________||
class All(Base):
    """select events that meet all conditions

    """

    def __init__(self, name='All', selections=[ ]):
        if name is None:
            name = 'All'
        super(All, self).__init__(name, selections)

    def __call__(self, event):
        for s in self.selections:
            if not s(event):
                return False
        return True

##__________________________________________________________________||
class Any(Base):
    """select events that meet any of the conditions

    """

    def __init__(self, name='Any', selections=[ ]):
        if name is None:
            name = 'Any'
        super(Any, self).__init__(name, selections)

    def __call__(self, event):
        for s in self.selections:
            if s(event):
                return True
        return False

##__________________________________________________________________||
class Not(object):
    """select events that do NOT pass the selection

    """

    def __init__(self, selection, name='Not'):
        if name is None:
            name = 'Not'
        self.name = name
        self.selection = selection

    def __repr__(self):
        return '{}(name={!r}, selection={!r})'.format(
            self.__class__.__name__,
            self.name,
            self.selection
        )

    def begin(self, event):
        if hasattr(self.selection, 'begin'):
            self.selection.begin(event)

    def __call__(self, event):
        return not self.selection(event)

    def event(self, event):
        return self(event)

    def end(self):
        if hasattr(self.selection, 'end'):
            self.selection.end()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import time

import ipywidgets as widgets
from IPython.display import display

##__________________________________________________________________||
class ProgressBarJupyter(object):
    def __init__(self):
        self.interval = 0.1 # [second]

        self.container_widget = None
        self.active_box_list = [ ]
        self.complete_box_list = [ ]
        self.widget_dict = { } # {taskid: [box, bar, label]}

        self._read_time()

    def __repr__(self):
        return '{}()'.format(
            self.__class__.__name__
        )

    def nreports(self):
        return len(self.active_box_list)

    def present(self, report):

        if not self._need_to_update(report):
            return

        if self.container_widget is None:
            self.container_widget = widgets.VBox()
            display(self.container_widget)

        if report.taskid not in self.widget_dict:
            self._create_widget(report)

        self._update_widget(report)

        self._reorder_widgets(report)

        self._read_time()

    def _create_widget(self, report):
        bar = widgets.IntProgress(
            value=report.done, min=0, max=report.total,
            description='',
            bar_style='', # 'success', 'info', 'warning', 'danger' or ''
            orientation='horizontal'
         )
        label = widgets.HTML(value='')
        box = widgets.HBox([bar, label])
        self.active_box_list.append(box)
        self.container_widget.children = self.complete_box_list + self.active_box_list
        self.widget_dict[report.taskid] = [box, bar, label]

    def _update_widget(self, report):

        percent = float(report.done)/report.total if report.total > 0 else 1
        percent = round(percent * 100, 2)
        percent = '<pre>{:6.2f}%</pre>'.format(percent)

        box = self.widget_dict[report.taskid][0]

        bar = self.widget_dict[report.taskid][1]
        bar.value = report.done
        bar.max = report.total
        bar.description = percent
        if report.done >= report.total:
            bar.bar_style = 'success'

        label = self.widget_dict[report.taskid][2]
        name_field_length = 32
        percent = float(report.done)/report.total if report.total > 0 else 1
        bar = (':' * int(percent * 40)).ljust(40, " ")
        percent = round(percent * 100, 2)
        name = report.name[0:name_field_length]
        label.value = '<pre> | {:8d} / {:8d} |:  {:<{}s}</pre>'.format(report.done, report.total, name, name_field_length)

    def _reorder_widgets(self, report):
        if report.done >= report.total:
            box, bar, label = self.widget_dict[report.taskid]
            if box in self.active_box_list:
                self.active_box_list.remove(box)
                self.complete_box_list.append(box)
                self.container_widget.children = self.complete_box_list + self.active_box_list

    def _need_to_update(self, report):
        if self._time() - self.last_time > self.interval:
            return True
        if report.done == report.total:
            return True
        if report.done == 0:
            return True
        return False

    def _time(self):
        return time.time()

    def _read_time(self):
        self.last_time = self._time()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from alphatwirl.misc.deprecation import _renamed_class_method_option
from alphatwirl.progressbar import atpbar

##__________________________________________________________________||
class EventLoop(object):
    """An event loop

    Args:
        build_events: A picklable function to create events.
        reader: An event reader. This must be picklable before
            `begin()` is called and after `end` is called.
        progressbar_label (optional): a label shown by the progress
            bar

    """
    @_renamed_class_method_option(old='name', new='progressbar_label')
    def __init__(self, build_events, reader, progressbar_label=None):
        self.build_events = build_events
        self.reader = reader

        ##
        if progressbar_label is None:
            self.progressbar_label = self.__class__.__name__
        else:
            self.progressbar_label = progressbar_label

        ##
        name_value_pairs = (
            ('build_events', self.build_events),
            ('reader', self.reader),
            ('progressbar_label', self.progressbar_label),
        )
        self._repr = '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]))

    def __repr__(self):
        return self._repr

    def __call__(self):
        events = self.build_events()
        self.reader.begin(events)
        for event in atpbar(events, name=self.progressbar_label):
            self.reader.event(event)
        self.reader.end()
        return self.reader

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import gzip
import shutil

try:
   import cPickle as pickle
except:
   import pickle

import pytest

import alphatwirl
import alphatwirl.concurrently
from alphatwirl.concurrently import TaskPackage

##__________________________________________________________________||
@pytest.fixture()
def env():
   # add path to alphatwirl to PYTHONPATH in env for subprocess
   env = os.environ.copy()
   alphatwirl_path = os.path.dirname(os.path.dirname(os.path.abspath(alphatwirl.__file__)))
   if 'PYTHONPATH' in env:
      env['PYTHONPATH'] = '{}:{}'.format(alphatwirl_path, env['PYTHONPATH'])
   else:
      env['PYTHONPATH'] = alphatwirl_path
   return env

@pytest.fixture()
def workingarea_path(tmpdir_factory):
   ret = str(tmpdir_factory.mktemp(''))
   src = os.path.join(os.path.dirname(alphatwirl.concurrently.__file__), 'run.py')
   dest = os.path.join(ret, 'run.py')
   shutil.copy(src, dest)
   return ret

##__________________________________________________________________||
def task_null(*args, **kwargs):
   return

@pytest.fixture()
def package_rel_path_task_null(workingarea_path):

   ret = 'task_00009.p.gz'

   package = TaskPackage(
      task = task_null,
      args = [ ],
      kwargs =   { }
   )

   path = os.path.join(workingarea_path, ret)
   with gzip.open(path, 'wb') as f:
      pickle.dump(package, f, protocol = pickle.HIGHEST_PROTOCOL)

   return ret

@pytest.mark.script_launch_mode('subprocess')
def test_run_task_null(script_runner, workingarea_path, package_rel_path_task_null, env):

   script_path = os.path.join('.', 'run.py')
   args = [package_rel_path_task_null]
   ret = script_runner.run(script_path, *args, cwd=workingarea_path, env=env)
   assert ret.success
   assert '' == ret.stderr
   assert '' == ret.stdout

##__________________________________________________________________||
def task_raise(*args, **kwargs):
    raise Exception('test error from task_raise()')

@pytest.fixture()
def package_rel_path_task_raise(workingarea_path):

   ret = 'task_00009.p.gz'

   package = TaskPackage(
      task = task_raise,
      args = [ ],
      kwargs =   { }
   )

   path = os.path.join(workingarea_path, ret)
   with gzip.open(path, 'wb') as f:
      pickle.dump(package, f, protocol = pickle.HIGHEST_PROTOCOL)

   return ret

@pytest.mark.script_launch_mode('subprocess')
def test_run_task_null(script_runner, workingarea_path, package_rel_path_task_raise, env):

   script_path = os.path.join('.', 'run.py')
   args = [package_rel_path_task_raise]
   ret = script_runner.run(script_path, *args, cwd=workingarea_path, env=env)
   assert ret.success == False
   assert 'Exception:' in ret.stderr
   assert '' == ret.stdout

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class Parallel(object):
    def __init__(self, progressMonitor, communicationChannel, workingarea=None):
        self.progressMonitor = progressMonitor
        self.communicationChannel = communicationChannel
        self.workingarea = workingarea

    def __repr__(self):
        name_value_pairs = (
            ('progressMonitor',      self.progressMonitor),
            ('communicationChannel', self.communicationChannel),
            ('workingarea',          self.workingarea)

        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def begin(self):
        self.progressMonitor.begin()
        self.communicationChannel.begin()

    def terminate(self):
        self.communicationChannel.terminate()

    def end(self):
        self.progressMonitor.end()
        self.communicationChannel.end()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class EventLoopRunner(object):
    """This class runs instances of `EventLoop` and keeps the results. It
    will return the results when `end()` is called.

    """
    def __init__(self):
        self.idx = -1 # so it starts from 0
        self.idx_result_pairs = [ ]

    def __repr__(self):
        return '{}()'.format(
            self.__class__.__name__,
        )

    def begin(self):
        self.idx_result_pairs = [ ]

    def run(self, eventLoop):
        self.idx += 1
        result = eventLoop()
        self.idx_result_pairs.append((self.idx, result))
        return self.idx

    def run_multiple(self, eventLoops):
        idxs = [ ]
        for eventLoop in eventLoops:
            idxs.append(self.run(eventLoop))
        return idxs

    def poll(self):
        return self.receive()

    def receive_one(self):
        if self.idx_result_pairs:
            return self.idx_result_pairs.pop(0)
        return None

    def receive(self):
        ret = self.idx_result_pairs[:]
        del self.idx_result_pairs[:]
        return ret

    def end(self):
        return [r for _, r in self.receive()]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from alphatwirl.selection.modules.LambdaStr import LambdaStr
import unittest

##__________________________________________________________________||
class MockEvent(object): pass

##__________________________________________________________________||
class Test_LambdaStr(unittest.TestCase):

    def test_standard(self):
        obj = LambdaStr(lambda_str = 'ev : True', name = 'test_true')

        event = MockEvent()
        obj.begin(event)
        self.assertTrue(callable(obj.func))

        event = MockEvent()
        self.assertTrue(obj(event))

        event = MockEvent()
        self.assertTrue(obj.event(event))

        obj.end()
        self.assertIsNone(obj.func)


    def test_raise(self):
        obj = LambdaStr(lambda_str = 'ev: aaa', name = 'test_true')

        event = MockEvent()
        obj.begin(event)

        event = MockEvent()
        self.assertRaises(ValueError, obj, event)

        obj.end()


##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class Collector(object):

    """This class collects results, i.e., this class combines results
    of readers and deliver them.

    Methods for combination and delivery are specified at the
    instantiation.

    Readers are typically instances of the same class initialized in
    the same way. Each reader reads a data set..

    The method ``collect`` is called with a list of pairs of a data
    set and a reader after the event loop. It returns the combined
    results.

    """

    def __init__(self, resultsCombinationMethod, deliveryMethod = None):
        self.resultsCombinationMethod = resultsCombinationMethod
        self.deliveryMethod = deliveryMethod if deliveryMethod is not None else NullDeliveryMethod()

        self._repr_pairs = [
            ('resultsCombinationMethod', self.resultsCombinationMethod),
            ('deliveryMethod',           self.deliveryMethod),
        ]

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in self._repr_pairs]),
        )

    def __str__(self):
        nwidth = max(len(n) for n, _ in self._repr_pairs)
        nwidth += 4
        return '{}:\n{}'.format(
            self.__class__.__name__,
            '\n'.join(['{:>{}}: {!r}'.format(n, nwidth, v) for n, v in self._repr_pairs]),
        )

    def collect(self, dataset_readers_list):
        results = self.resultsCombinationMethod.combine(dataset_readers_list)
        self.deliveryMethod.deliver(results)
        return results

##__________________________________________________________________||
class NullDeliveryMethod(object):
    def deliver(self, results): pass

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.selection import build_selection
from alphatwirl.selection.modules import All
from alphatwirl.selection.modules import Any
from alphatwirl.selection.modules import Not
from alphatwirl.selection.modules.LambdaStr import LambdaStr

##__________________________________________________________________||
@pytest.fixture()
def MockFactoryDispatcher(monkeypatch):
    ret = mock.Mock()
    module = sys.modules['alphatwirl.selection.funcs']
    monkeypatch.setattr(module, 'FactoryDispatcher', ret)
    return ret

def test_call_kargs(MockFactoryDispatcher):

    kargs = dict(
        arg1=10,
        arg2=20,
        level=dict(factory='test_level1', arg2=2, arg3=3)
    )

    obj = build_selection(**kargs)

    assert MockFactoryDispatcher() is obj

    expected = [mock.call(
        AllClass=All, AnyClass=Any, NotClass=Not,
        LambdaStrClass=LambdaStr, aliasDict={ },
        arg1=10, arg2=20, level=dict(factory='test_level1', arg2=2, arg3=3))
    ] == MockFactoryDispatcher.call_args_list

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from alphatwirl.misc.removal import _removed

##__________________________________________________________________||
@_removed(msg='use alphatwirl.parallel.build.build_parallel() instead.')
def build_progressMonitor_communicationChannel(quiet, processes):
    pass

##__________________________________________________________________||
